/*****************************************************************************
 *
 * wave.c
 *
 * PHASEX:  [P]hase [H]armonic [A]dvanced [S]ynthesis [EX]periment
 *
 * Copyright (C) 1999-2007 William Weston <weston@sysex.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************/
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include "phasex.h"
#include "wave.h"
#include "filter.h"
#include "jack.h"


/* base tuning frequency -- can be overridden on command line */
double		a4freq = A4FREQ;

/* this is _the_ wave table */
float_t		wave_table[NUM_WAVEFORMS][WAVEFORM_SIZE];

/* frequency table for midi notes */
float_t		freq_table[128][648];

/* lookup table for exponential frequency scaling */
float_t		freq_shift_table[FREQ_SHIFT_TABLE_SIZE];

/* env table for envelope durations in samples */
int		env_table[128];

/* env curve for envelope values to actual amplitudes */
float_t		env_curve[ENV_CURVE_SIZE+1];

/* table for mixing gain, w/ unity at cc=127 */
float_t		mix_table[128];

/* table for general purpose gain, w/ unity at cc=120 */
float_t		gain_table[128];

/* keyfollow table for ctlr/key combo --> amplitude coefficient */
float_t		keyfollow_table[128][128];


/*****************************************************************************
 *
 * BUILD_FREQ_TABLE()
 *
 * Generates the (midi_note - 128) to frequency lookup table
 *
 *****************************************************************************/
void
build_freq_table(void) {
    double	basefreq;
    double	freq;
    double	halfstep;
    double	tunestep;
    int		octave;
    int		note;
    int		j;

    /* build table with 53 1/3 octaves, w/ MIDI range in center */

    /* normal halfstep is 12th of an even tempered octave */
    halfstep = exp (log (2.0) / 12.0);

    /* tuning works in 120th of a halfstep increments */
    tunestep = exp (log (2.0) / (12.0 * F_TUNING_RESOLUTION));

    /* for every master tune adjustment */
    for (j = 0; j < 128; j++) {
	basefreq = a4freq * pow (tunestep, (double)(j - 64));

	/* build the table with a freshly calculated base freq every octave */
	for (octave = 0; octave < 54; octave++) {

	    /* A4 down 27 octaves and 1 half-step */
	    /* (table starts at 256 half-steps below midi range) */
	    freq = basefreq * pow(2.0, octave - 27) / halfstep;

	    /* calculate notes for each octave with successive multiplications */
	    for (note = octave * 12; note < (octave + 1) * 12; note++) {
		freq_table[j][note] = (float_t)freq;
		freq                *= halfstep;
	    }
	}
    }
}


/*****************************************************************************
 *
 * BUILD_FREQ_SHIFT_TABLE()
 *
 * Generates lookup table for exponential function used for
 * fine frequency adjustments.
 *
 *****************************************************************************/
void
build_freq_shift_table(void) {
    double	halfstep;
    double	tunestep;
    int		index;
    int		octaves;
    int		halfsteps;
    int		tunesteps;


    halfstep  = exp (log (2.0) / 12.0);
    tunestep = exp (log (2.0) / (12.0 * F_TUNING_RESOLUTION));

    for (index = 0; index < FREQ_SHIFT_TABLE_SIZE / 2; index++) {
	tunesteps = (FREQ_SHIFT_TABLE_SIZE / 2) - index;
	halfsteps = tunesteps / TUNING_RESOLUTION;
	tunesteps %= TUNING_RESOLUTION;
	octaves   = halfsteps / 12;
	halfsteps %= 12;
	freq_shift_table[index] = (float_t)(1.0 /
	    (pow(2.0,      (double)octaves) *
	     pow(halfstep, (double)halfsteps) *
	     pow(tunestep, (double)tunesteps)));
    }
    freq_shift_table[(FREQ_SHIFT_TABLE_SIZE / 2)] = 1.0;
    for (index = (FREQ_SHIFT_TABLE_SIZE / 2) + 1; index < FREQ_SHIFT_TABLE_SIZE; index++) {
	tunesteps = index - (FREQ_SHIFT_TABLE_SIZE / 2);
	halfsteps = tunesteps / TUNING_RESOLUTION;
	tunesteps %= TUNING_RESOLUTION;
	octaves   = halfsteps / 12;
	halfsteps %= 12;
	freq_shift_table[index] = (float_t)
	    (pow(2.0,      (double)octaves) *
	     pow(halfstep, (double)halfsteps) *
	     pow(tunestep, (double)tunesteps));
    }
}


/*****************************************************************************
 *
 * BUILD_ENV_TABLES()
 *
 * Generates lookup tables for cc to envelope duration and
 * amplitude curve.
 *
 *****************************************************************************/
void
build_env_tables(void) {
    float_t	env;
    float_t	logval;
    float_t	logstep;
    int		j;

    /* envelope duration table */
    for (j = 0; j < 128; j++) {
	env_table[j] = ((j + 0) * (j + 0) * (sample_rate / 6000)) + 2;
    }

    /* envelope logarithmic volume curve */
    logval  = logf (16384.0);
    logstep = F_ENV_CURVE_SIZE * 2.0;
    env     = 1.0;
    for (j = ENV_CURVE_SIZE; j > 0; j--) {
	env_curve[j] = (float_t)(env);
	env          /= expf (logval / logstep);
	logstep      -= 1.0;
    }
    env_curve[0] = 0.0;
}


/*****************************************************************************
 *
 * BUILD_KEYFOLLOW_TABLE()
 *
 * Generates lookup tables for cc to keyfollow volume adjustment
 *
 *****************************************************************************/
void
build_keyfollow_table(void) {
    int		ctlr;
    int		key;
    float_t	slope;

    /* first the low freq weighted slope */
    for (ctlr = 0; ctlr < 64; ctlr++) {
	slope = (float_t)(64 - ctlr) / 64.0;
	for (key = 0; key < 128; key++) {
	    keyfollow_table[ctlr][key] = 1.0 - (slope * (((float_t)(key)) / 127.0));
	}
    }

    /* next the median */
    for (key = 0; key < 128; key++) {
	keyfollow_table[64][key] = 1.0;
    }

    /* last the high freq weighted slope */
    for (ctlr = 65; ctlr < 128; ctlr++) {
	slope = (float_t)(ctlr - 64) / 63.0;
	for (key = 0; key < 128; key++) {
	    keyfollow_table[ctlr][key] = 1.0 - (slope * (((float_t)(127 - key)) / 127.0));
	}
    }
}


/*****************************************************************************
 *
 * BUILD_MIX_TABLE()
 *
 * Generates lookup tables for cc to mix scalar mappings.
 *
 *****************************************************************************/
void
build_mix_table(void) {
    int		j;


    mix_table[0]   = 0.0;
    mix_table[127] = 1.0;
    for (j = 1; j < 127; j++) {
    	mix_table[j] = ((float_t)j) / 127.0;
    	mix_table[j] += sqrt(((float_t)j) / 127.0);
	mix_table[j] *= 0.5;
    }
}


/*****************************************************************************
 *
 * BUILD_GAIN_TABLE()
 *
 * Generates lookup tables for cc to gain scalar mappings.
 *
 *****************************************************************************/
void
build_gain_table(void) {
    int		j;
    float_t	gain;
    float_t	step;


    /* 1/2 dB steps */
    step = exp(log(10.0)/40.0);

    gain = 1.0;
    for (j = 119; j >= 0; j--) {
	gain /= step;
	gain_table[j] = gain;
    }
    gain_table[120] = 1.0;
    gain = 1.0;
    for (j = 121; j < 128; j++) {
	gain *= step;
	gain_table[j] = gain;
    }
}


/*****************************************************************************
 *
 * BUILD_WAVE_TABLES()
 *
 * Generates high res wave tables for use by all oscillators and lfos
 *
 *****************************************************************************/
void
build_waveform_tables(void) {
    int			sample_num;
    float_t		angle;


    for (sample_num = 0; sample_num < WAVEFORM_SIZE; sample_num++) {

	angle = 2 * M_PI * (float_t)sample_num / F_WAVEFORM_SIZE;

	wave_table[WAVE_SINE][sample_num]	= sin (angle);
	wave_table[WAVE_TRIANGLE][sample_num]	= func_triangle (sample_num);
	wave_table[WAVE_SAW][sample_num]	= func_saw (sample_num);
	wave_table[WAVE_REVSAW][sample_num]	= func_revsaw (sample_num);
	wave_table[WAVE_SQUARE][sample_num]	= func_square (sample_num);
	wave_table[WAVE_STAIR][sample_num]	= func_stair (sample_num);
	wave_table[WAVE_SAW_S][sample_num]	= func_saw (sample_num);
	wave_table[WAVE_REVSAW_S][sample_num]	= func_revsaw (sample_num);
	wave_table[WAVE_SQUARE_S][sample_num]	= func_square (sample_num);
	wave_table[WAVE_STAIR_S][sample_num]	= func_stair (sample_num);
	wave_table[WAVE_IDENTITY][sample_num]	= 1.0;
	wave_table[WAVE_NULL][sample_num]	= 0.0;
#if (NUM_WAVEFORMS >= 24)
	if (debug) {
	    printf("Warning: building %d wave tables\n", NUM_WAVEFORMS);
	}
	wave_table[WAVE_12][sample_num]		= sin (angle) * sin (angle * 2.0);
	wave_table[WAVE_13][sample_num]		= sin (angle) * sin (angle * 3.0);
	wave_table[WAVE_14][sample_num]		= sin (angle) * sin (angle * 4.0);
	wave_table[WAVE_15][sample_num]		= sin (angle) * sin (angle * 5.0);
	wave_table[WAVE_16][sample_num]		= sin (angle) * sin (angle * 6.0);
	wave_table[WAVE_17][sample_num]		= sin (angle) * sin (angle * 7.0);
	wave_table[WAVE_18][sample_num]		= sin (angle) * sin (angle * 8.0);
	wave_table[WAVE_19][sample_num]		= sin (angle) * sin (angle * 9.0);
	wave_table[WAVE_20][sample_num]		= sin (angle) * sin (angle * 2.0) * sin (angle * 3.0);
	wave_table[WAVE_21][sample_num]		= sin (angle) * sin (angle * 2.0) * sin (angle * 5.0);
	wave_table[WAVE_22][sample_num]		= sin (angle) * sin (angle * 2.0) * sin (angle * 7.0);
	wave_table[WAVE_23][sample_num]		= sin (angle) * sin (angle * 3.0) * sin (angle * 7.0);
#endif
    }

    /* bandlimit the _S (or BL) waveshapes */
    filter_wave_table (WAVE_SAW_S);
    filter_wave_table (WAVE_REVSAW_S);
    filter_wave_table (WAVE_SQUARE_S);
    filter_wave_table (WAVE_STAIR_S);
}

#if 0
/*****************************************************************************
 *
 * MASSAGE_WAVES()
 *
 * Smooths out vertical edges and quick reversals in wave table.
 * This is old code from 0.9.x, and sounds different than bandlimiting.
 *
 *****************************************************************************/
void
massage_waves(void) {
    float_t		delta = 100.0 / F_WAVEFORM_SIZE;
    float_t		average = 0.0;
    float_t		len = 2048.0;
    float_t		sublen = 2047.0;
    int			wave;
    int			j;

    /* Re-slope vertical edges on smoothed square and stair */
    for (j = 0; j < (WAVEFORM_SIZE / 100); j++) {
	/* square wave */
	wave_table[WAVE_SQUARE_S][j]                       = ((float_t)j * delta * 2.0) - 1.0;
	wave_table[WAVE_SQUARE_S][(WAVEFORM_SIZE / 2) + j] = 1.0 - ((float_t)j * delta * 2.0);

	/* stair step wave */
	wave_table[WAVE_STAIR_S][j]				= -1.0 + (float_t)j * delta;
	wave_table[WAVE_STAIR_S][(WAVEFORM_SIZE / 4) + j]	=  0.0 + (float_t)j * delta;
	wave_table[WAVE_STAIR_S][(WAVEFORM_SIZE / 2) + j]	=  1.0 - (float_t)j * delta;
	wave_table[WAVE_STAIR_S][(3 * WAVEFORM_SIZE / 4) + j]	=  0.0 - (float_t)j * delta;
    }

    /* Run forward and reverse sweep decayed averaging for smoothing */
    for (wave = WAVE_SAW_S; wave <= WAVE_STAIR_S; wave++) {
	/* don't do anything to null or identity waves */
	if ((wave == WAVE_IDENTITY) || (wave == WAVE_NULL))
	    continue;
	/* Get average of last 1024 samples */
	average = 0.0;
	for (j = WAVEFORM_SIZE - 1024; j < WAVEFORM_SIZE; j++) {
	    average += wave_table[wave][j];
	}
	average = average / len;
	/* forward sweep */
	for (j = 0; j < WAVEFORM_SIZE; j++) {
	    average = (average * sublen + wave_table[wave][j]) / len;
	    wave_table[wave][j] = average;
	}
	/* Get average of first 2048 samples */
	average = 0.0;
	for (j = 0; j < 1024; j++) {
	    average += wave_table[wave][j];
	}
	average = average / len;
	/* reverse sweep */
	for (j = WAVEFORM_SIZE - 1; j >= 0; j--) {
	    average = (average * sublen + wave_table[wave][j]) / len;
	    wave_table[wave][j] = average;
	}
    }
}
#endif


/*****************************************************************************
 *
 * Functions for the generating the waveform samples
 * The synth engine should never use these directly due to overhead.
 * Use the wave table instead!
 *
 *****************************************************************************/


/*****************************************************************************
 *
 * FUNC_SQUARE()
 *
 * Square function for building wave table
 *
 *****************************************************************************/
float_t
func_square(unsigned int sample_num) {
    if (sample_num < (WAVEFORM_SIZE / 2))
      return 1.0;
    return -1.0;
}


/*****************************************************************************
 *
 * FUNC_Saw()
 *
 * Saw function for building wave table
 *
 *****************************************************************************/
float_t
func_saw(unsigned int sample_num) {
    return (2.0 * (float_t)sample_num / F_WAVEFORM_SIZE) - 1.0;
}


/*****************************************************************************
 *
 * FUNC_REVSAW()
 *
 * Reverse saw wave function for building wave table.
 *
 *****************************************************************************/
float_t
func_revsaw(unsigned int sample_num) {
    return (2.0 * (F_WAVEFORM_SIZE - (float_t)sample_num) / F_WAVEFORM_SIZE) - 1.0;
}


/*****************************************************************************
 *
 * FUNC_SAW_S()
 *
 * Smoothed saw wave function for building wave table.
 *
 *****************************************************************************/
float_t
func_saw_s(unsigned int sample_num) {
    int		threshold = WAVEFORM_SIZE - 96;

    if (sample_num < threshold)
	return (2.0 * (float_t)sample_num / (float_t)threshold) - 1.0;
    else
	return 1.0 - (2.0 * ((float_t)sample_num - (float_t)threshold)
		      / (F_WAVEFORM_SIZE - (float_t)threshold));
}


/*****************************************************************************
 *
 * FUNC_REVSAW_S()
 *
 * Smoothed reverse saw wave function for building wave table.
 *
 *****************************************************************************/
float_t
func_revsaw_s(unsigned int sample_num) {
    int		threshold = 96;

    if (sample_num < threshold)
	return (2.0 * (float_t)sample_num / (float_t)threshold) - 1.0;
    else
	return 1.0 - (2.0 * ((float_t)sample_num - (float_t)threshold)
		      / (F_WAVEFORM_SIZE - (float_t)threshold));
}


/*****************************************************************************
 *
 * FUNC_TRIANGLE()
 *
 * Triangle wave function for building wave table
 *
 *****************************************************************************/
float_t
func_triangle(unsigned int sample_num) {
    if (sample_num < (WAVEFORM_SIZE / 4))
	return 4.0 * (float_t)sample_num / F_WAVEFORM_SIZE;
    sample_num -= (WAVEFORM_SIZE / 4);
    if (sample_num < (WAVEFORM_SIZE / 4))
	return 4.0 * ((F_WAVEFORM_SIZE / 4.0) - (float_t)sample_num) / F_WAVEFORM_SIZE;
    sample_num -= (WAVEFORM_SIZE / 4);
    if (sample_num < (WAVEFORM_SIZE / 4))
	return -4.0 * (float_t)sample_num / F_WAVEFORM_SIZE;
    sample_num -= (WAVEFORM_SIZE / 4);
    return -4.0 * ((F_WAVEFORM_SIZE / 4.0) - (float_t)sample_num) / F_WAVEFORM_SIZE;
}


/*****************************************************************************
 *
 * FUNC_STAIR()
 *
 * Stairstep wave function for building wave table
 *
 *****************************************************************************/
float_t
func_stair(unsigned int sample_num) {
    if (sample_num < (WAVEFORM_SIZE / 4))
	return 0.0;
    sample_num -= (WAVEFORM_SIZE / 4);
    if (sample_num < (WAVEFORM_SIZE / 4))
	return 1.0;
    sample_num -= (WAVEFORM_SIZE / 4);
    if (sample_num < (WAVEFORM_SIZE / 4))
	return 0.0;
    return -1.0;
}
